use crate::handlers;
use crate::state::AppState;
use crate::ui::UIMessage;
use axum::{routing, Router};
use serde::{Deserialize, Serialize};
use std::net::SocketAddr;
use tokio::{net::TcpListener, sync::mpsc};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ServerSettings {
    pub port: String,
}

#[derive(Debug, Clone)]
pub enum ServerMessage {
    UpdateSettings(ServerSettings),
    Interact,
    Shutdown,
}

pub async fn run_server(
    app_state: AppState,
    mut rx: mpsc::Receiver<ServerMessage>,
    tx: mpsc::Sender<UIMessage>,
) {
    loop {
        let should_run = {
            let guard = app_state.should_run.lock().unwrap();
            *guard
        };

        if !should_run {
            eprintln!("Should run is false.");
            break;
        }

        let addr = {
            let settings = app_state.settings.lock().unwrap();
            let port: u16 = settings.port.parse().unwrap_or(3311);
            SocketAddr::from(([0, 0, 0, 0], port))
        };

        let server_app_state = app_state.clone();

        println!("listening on {}", addr);

        let server_handle = tokio::spawn(async move {
            let app = Router::new()
                .route("/", routing::get(handlers::home))
                .route("/play/:hash", routing::get(handlers::play_handler))
                .route("/stop", routing::get(handlers::stop_handler))
                .with_state(server_app_state.clone());
            let listener = TcpListener::bind(&addr).await.unwrap();
            axum::serve(listener, app.into_make_service()).await
        });
        let server_abort_handle = server_handle.abort_handle();

        tokio::select! {
            _ = server_handle => {
                // Server has stopped
                eprintln!("Server stopped.");
                let mut should_run_guard = app_state.should_run.lock().unwrap();
                *should_run_guard = false;
                drop(should_run_guard);
                break;
            },
            Some(message) = rx.recv() => {
                match message {
                    ServerMessage::UpdateSettings(new_settings) => {
                        println!("Received new settings: {:?}", new_settings);
                        let mut settings_guard = app_state.settings.lock().unwrap();
                        *settings_guard = new_settings;
                        drop(settings_guard); // Ensure the lock is released before starting the new server
                        // Restart the server with new settings
                        println!("Aborting current server...");
                        server_abort_handle.abort();
                    },
                    ServerMessage::Shutdown => {
                        // Handle server shutdown
                        println!("Shutting down server...");
                        let mut should_run_guard = app_state.should_run.lock().unwrap();
                        *should_run_guard = false;
                        drop(should_run_guard);
                        server_abort_handle.abort();
                        println!("Server shutdown.");
                        break;
                    },
                    ServerMessage::Interact => {
                        // Example interaction: print current settings
                        let settings_guard = app_state.settings.lock().unwrap();
                        println!("Current settings: {:?}", *settings_guard);
                        server_abort_handle.abort();
                    },
                }
            }
        }
    }
    tx.try_send(UIMessage::ServerHasQuit).unwrap_or_default();
    eprintln!("Exiting run_server.");
}
